==> mkreadmes-1.1.2.tar.bz2: mkreadmes-1.1.2/index.c <==
/*
 *	index.c
 *
 *	routines that deal with the index file
 *
 *	$Id: index.c,v 1.67 2012/03/07 17:11:17 conrads Exp $
 *
 *****************************************************************************/

#include <fcntl.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/utsname.h>
#include <unistd.h>

#include "mkreadmes.h"

char	*index_field[NUMINDEXFIELDS];	/* array of pointers to the index fields
											for the	current port
										 */
char	*index_buf = NULL;			/* pointer to the index file buffer */

char	**index_categories = NULL;	/* pointer to an array of pointers to the
										first line of each category's section
										within the index buffer
									 */

char	*next_index;	/* pointer to the next index line to search from */
char	*next_cat;		/* pointer to the first index line of next category */

extern	char	ports_dir[PATH_MAX];	/* top-level directory */

/* prototypes for routines in other source files */

extern	char   *load_file(const char *path);						/* init.c */
extern	char   *skip_leading(const char* str, const char* pattern); /* init.c */

/*****************************************************************************/

static void split_index_line(const char *line)
/*
 *	convert the index line pointed to by 'line' into an array of nul-terminated
 *	strings stored in index_line[], saving pointers to each field in the global
 *	index_field[] array
 */
{
	static	char	index_line[MAXINDEXLINE];	/* list of index fields for the
													 latest port found by
													 search_index()
												 */
	char   *ptr;
	int		i = 0;

	/* we DO NOT want to pass 'line' directly to strsep(), as doing so will
		trash the index buffer, so copy it first, and pass a pointer to the
		copy instead

		we also can't just pass index_line[] to strsep(), as it's not a
		compatible type, so we need to use a pointer
	 */

	ptr = strcpy(index_line, line);

	while (ptr)
	{
		index_field[i++] = strsep(&ptr, INDEXSEPARATOR);
	}
}

/*****************************************************************************/

int search_index(const char *path)
/*
	Find a port's entry in the index buffer by first converting its real path to
	the canonical form used in the index file's PORT field, then comparing it to
	the PORT index field, and if found, convert its index fields into an array
	of strings stored in index_line[], saving pointers to each field in the
	index_field[] array

	On entry, the global variable 'next_index' will already be pointing to the
	first line in the index buffer to check, and next_cat will be pointing to
	the first index line of the next category (the search cutoff point)

	Returns 1 if found, 0 otherwise
*/
{
	char	port_path[PATH_MAX] = PORTSTOP;	/* the canonical "/usr/ports" */

	char   *port;		/* pointer to PORT field in current index line */
	int		len;		/* length of PORT field */

	char   *index_ptr = next_index;	/* pointer into the index buffer */

	/* scan the index file buffer one line at a time, starting at *next_index,
		comparing the port's path against the current index line's PORT field

		the search will terminate with failure if index_ptr runs into the start
		of the next category's section in the index buffer (or the end of the
		buffer, if we're searching within the last category in the index)
	 */

	/* append category/port from path to PORTSTOP in port_path */
	strcat(port_path, skip_leading(path, ports_dir));

	while (index_ptr != next_cat)	/* only search one category */
	{
		/* skip past the first INDEXSEPARATOR and set 'port' to point to the
			PORT field
		 */
		port = index_ptr + strcspn(index_ptr, INDEXSEPARATOR) + 1;

		/* get length of PORT field */
		len = strcspn(port, INDEXSEPARATOR);

		/* compare port_path to the PORT field */
		if ((len == strlen(port_path)) && (strncmp(port, port_path, len) == 0))
		{
			/* we have a match

				break the index line into its individual fields for later use,
				storing pointers to each field in the global index_field[] array
			 */
			split_index_line(index_ptr);

			/* bump the next_index pointer to the next line in the index buffer,
				for any subsequent searches within the same category
			 */
			next_index = index_ptr + strlen(index_ptr) + 1;
			return 1;
		}

		index_ptr += strlen(index_ptr) + 1;	/* bump pointer to next line and
												try again */
	}

	/* falling through to here means no match found for this port */

	return 0;
}

/*****************************************************************************/

static char *get_index_line_category(char category[], char *line)
/*
	find the category field within an index line, chop off all but the first
	category name, and copy it into category[]

	returns a pointer to category
*/
{
	int	field, len;

	/* locate the category field */

	for (field = PKG; field < CATEGORY; ++field)
	{
		line += strcspn(line, INDEXSEPARATOR) + 1;
	}

	/* OK, now we're pointing at the first character of the CATEGORY field;
		split off all but the first category name */

	len = strcspn(line, CATSEPARATOR);	/* split on space or INDEXSEPARATOR */
	strncpy(category, line, len);	/* copy up to the space or INDEXSEPARATOR */
	category[len] = '\0';			/* and nul-terminate the category */

	return category;
}

/*****************************************************************************/

static int validate_index_line(const char *line)
/*
	Make sure all of the fields we need in an index line actually contain data

	Returns 1 if line is valid, 0 if any of the tested fields are empty
 */
{
	split_index_line(line);

	if ((strlen(index_field[PKG]) == 0) ||
		(strlen(index_field[PORT]) == 0) ||
		(strlen(index_field[COMMENT]) == 0) ||
		(strlen(index_field[DESCR]) == 0) ||
		(strlen(index_field[EMAIL]) == 0) ||
		(strlen(index_field[CATEGORY]) == 0))
		return 0;

	return 1;
}

/*****************************************************************************/

static int setup_index_categories(void)
/*
	setup an array of pointers, cross-indexed against the categories list,
	to the first line of each category's section within the index buffer.

	this will allow us to easily find the start of a category within the index
	buffer by name, and speed up port searches quite dramatically by limiting
	our searches to only the category to which a given port	belongs.

	on entry, the index buffer will contain a series of lines, each terminated
	with a newline character, with a single nul end-of-file marker

	this routine will replace each newline with a nul character, transforming
	the buffer into an array of strings, with a single extra nul end-of-file
	marker at the end

	each index line will be validated as well, to ensure all of the needed
	fields actually contain data, returning an error if an improperly formatted
	line is encountered

	returns 0 on success, -1 on error
*/
{
	char   *index_ptr = index_buf;		/* pointer into index_buf */
	char	this_category[MAXCATNAME];	/* current category */
	char	next_category[MAXCATNAME];	/* next category */
	int		i = 0;					/* index into index_categories[] array */
	extern	int	num_categories;		/* number of categories found in top-level
										Makefile
									 */
	int		len;					/* length of current index line */
	int		line_no = 1;			/* for reporting the line number of any
										improperly formatted index line
									 */

	/* allocate memory for the array where we'll save pointers to the first line
		of each category's section within the index buffer
	 */
	if ((index_categories = calloc((num_categories + 1), sizeof(char *))) == NULL)
	{
		perror("calloc() failure");
		return -1;
	}

	/* scan the index buffer one newline-terminated line at a time,
		replacing the newline with nul.

		whenever a new category is found, add a pointer to the line
		containing it to the index_categories[] array
	 */

	while (*index_ptr)
	{
		/* whenever we're at the top of this loop, index_ptr will be pointing
			at the start of a new category section in the index buffer (or end
			of buffer)
		 */

		/* replace newline with nul */
		index_ptr[len = strcspn(index_ptr, "\n")]= '\0';

		/* Make sure the line is properly formatted */
		if (!validate_index_line(index_ptr))
		{
			fprintf(stderr, "Error in index file line %d: line = \"%s\"\n",
				line_no, index_ptr);
			return -1;
		}

		index_categories[i++] = index_ptr;	/* save a pointer to the start of
												this category's section
											 */

		/* on all but the first pass, next_category will already contain
			the name of the new category, so we can avoid a needless
			duplication of effort here */
		if (index_ptr == index_buf)
			/* copy this line's category into this_category */
			get_index_line_category(this_category, index_ptr);
		else
			/* copy the category we already picked up on the last pass */
			strcpy(this_category, next_category);

		/* now keep bumping index_ptr to the next index line until we hit a
			new category, or reach end of file
		 */
		while (index_ptr += len + 1)
		{
			if (*index_ptr == '\0')
				break;					/* we've reached end of buffer */

			++line_no;

			/* replace newline with nul */
			index_ptr[len = strcspn(index_ptr, "\n")]= '\0';

			/* Make sure the line is properly formatted */
			if (!validate_index_line(index_ptr))
			{
				fprintf(stderr, "Error in index file line %d: line = \"%s\"\n",
					line_no, index_ptr);
				return -1;
			}

			/* get this line's category and compare it to current category */
			get_index_line_category(next_category, index_ptr);

			if (strcmp(this_category, next_category) != 0)
				break;	/* we've hit a new category,
							back to top of outer loop */
		}
	}

	index_categories[i] = NULL;	/* mark end of list */

	return 0;
}

/*****************************************************************************/

static int gen_index_path(char *index_path)
/*
 *	generate the fully qualified pathname to the index file,
 *	storing it back into index_path[]
 *
 *	returns 0 on success, or -1 on error
 */
{
	extern	char	ports_dir[PATH_MAX];	/* top-level ports directory */
	struct	utsname	uname_struct;		/* information returned by uname() */

	/* get the OS release number */

	if (uname(&uname_struct) == -1)
	{
		perror("uname()");
		return -1;
	}

	/* set index_path to ports_dir/INDEXNAME */

	strcpy(index_path, ports_dir);
	strcat(index_path, INDEXNAME);	/* index name minus OS release number */

	/* append release number to pathname */

	strncat(index_path, uname_struct.release,
		strspn(uname_struct.release, "0123456789"));

	return 0;
}

/*****************************************************************************/

int load_index(void)
/*
	Load the index file into a dynamically allocated buffer which will be used
	throughout the lifetime of the program.  This helps to ensure the fastest
	possible access to the file's data.

	The index file, as stored on disk, consists of a series of lines of data,
	one line per port, each line being terminated by a newline character (not
	nul).  Each individual line consists of 13 fields of data separated by the
	INDEXSEPARATOR character "|" (#defined in mkreadmes.h).  The last field of
	each line is followed immediately by the newline character (i.e., there is
	no INDEXSEPARATOR at end of line).

	The file will be stored in memory exactly as it is stored on disk, with the
	following two exceptions:

	The newline character terminating each line will be replaced with a nul,
	effectively turning each line into a standard nul-terminated string, and
	the index buffer into an array of strings

	A single extra nul character will be added at the end of the buffer (after
	the	nul character which terminates the last index line) as an end-of-file
	marker.

	A pointer to the file's location in memory will be stored in the global
	variable index_buf.

	Returns 0 on success, or -1 on failure
*/
{
	char	index_path[PATH_MAX];

	/* get the fully qualified pathname into index_path */
	if (gen_index_path(index_path) != 0)
		return -1;

	/* load the file into memory, saving its address in the global index_buf */
	if ((index_buf = load_file(index_path)) == NULL)
		return -1;

	/* setup an array of pointers, cross-indexed against the categories list,
		to the first line of each category's section within the index buffer,
		while also replacing each newline with a nul character, transforming
		the buffer into an array of strings with a single extra nul end-of-file
		marker at the end
	 */
	return setup_index_categories();
}

==> mkreadmes-1.2.tar.bz2: mkreadmes-1.2/index.c <==
/*
 *	index.c
 *
 *	routines that deal with the index file
 *
 *	$Id: index.c,v 1.76 2012/03/10 02:17:07 conrads Exp $
 *
 *****************************************************************************/

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/param.h>
#include <sys/stat.h>
#include <sys/utsname.h>

#include "mkreadmes.h"

char	*index_field[NUMINDEXFIELDS];	/* array of pointers to the index fields
											for the	current port
										 */
char	*index_buf = NULL;				/* pointer to the index file buffer */

extern	char *(*index_categories)[];	/* pointer to an array of pointers
											to the first line of each
											category's section within the
											index buffer
										 */
extern	char   *next_index;	/* pointer to the next index line to search from */
extern	char   *next_cat;	/* pointer to the first line of next category */

extern	char	ports_dir[PATH_MAX];	/* top-level directory */

extern	size_t	num_categories;		/* number of categories found in top-level
										Makefile
									 */
/* prototypes for global functions */

int		load_index(void);
int		search_index(const char *path);
int		set_index_categories(void);

/* prototypes for routines in other source files */

extern	char   *load_file(const char *path);						/* util.c */
extern	char   *skip_leading(const char* str, const char* pattern); /* util.c */

/*****************************************************************************/

static void split_index_line(const char *line)
/*
 *	convert the index line pointed to by 'line' into an array of nul-terminated
 *	strings stored in index_line[], saving pointers to each field in the global
 *	index_field[] array
 */
{
	static	char	index_line[MAXINDEXLINE];	/* list of index fields for the
													 latest port found by
													 search_index()
												 */
	char   *ptr;
	int		i = 0;

	/* we DO NOT want to pass 'line' directly to strsep(), as doing so will
		trash the index buffer, so copy it first, and pass a pointer to the
		copy instead

		we also can't just pass index_line[] to strsep(), as it's not a
		compatible type, so we need to use a pointer
	 */

	ptr = strcpy(index_line, line);

	while (ptr)
	{
		index_field[i++] = strsep(&ptr, FIELDSEPARATOR);
	}
}

/*****************************************************************************/

int search_index(const char *path)
/*
	Find a port's entry in the index buffer by first converting its real path to
	the canonical form used in the index file's PORT field, then comparing it to
	the PORT index field, and if found, convert its index fields into an array
	of strings stored in index_line[], saving pointers to each field in the
	index_field[] array

	On entry, the global variable 'next_index' will already be pointing to the
	first line in the index buffer to check, and next_cat will be pointing to
	the first index line of the next category (the search cutoff point)

	Returns 1 if found, 0 otherwise
*/
{
	char	port_path[PATH_MAX] = PORTSTOP;	/* the canonical "/usr/ports" */

	char   *port;		/* pointer to PORT field in current index line */
	size_t	len;		/* length of PORT field */

	char   *index_ptr = next_index;	/* pointer into the index buffer */

	/* scan the index file buffer one line at a time, starting at *next_index,
		comparing the port's path against the current index line's PORT field

		the search will terminate with failure if index_ptr runs into the start
		of the next category's section in the index buffer (or the end of the
		buffer, if we're searching within the last category in the index)
	 */

	/* append category/port from path to PORTSTOP in port_path */
	strcat(port_path, skip_leading(path, ports_dir));

	while (index_ptr != next_cat)	/* only search one category */
	{
		/* skip past the first FIELDSEPARATOR and set 'port' to point to the
			PORT field
		 */
		port = index_ptr + strcspn(index_ptr, FIELDSEPARATOR) + 1;

		/* get length of PORT field */
		len = strcspn(port, FIELDSEPARATOR);

		/* compare port_path to the PORT field */
		if ((len == strlen(port_path)) && (strncmp(port, port_path, len) == 0))
		{
			/* we have a match

				break the index line into its individual fields for later use,
				storing pointers to each field in the global index_field[] array
			 */
			split_index_line(index_ptr);

			/* bump the next_index pointer to the next line in the index buffer,
				for any subsequent searches within the same category
			 */
			next_index = index_ptr + strlen(index_ptr) + 1;
			return 1;
		}

		index_ptr += strlen(index_ptr) + 1;	/* bump pointer to next line and
												try again */
	}

	/* falling through to here means no match found for this port */

	return 0;
}

/*****************************************************************************/

static char *get_index_line_category(char category[], char *line)
/*
	find the category field within an index line, chop off all but the first
	category name, and copy it into category[]

	returns a pointer to category
*/
{
	int		field;
	size_t	len;

	/* locate the category field */

	for (field = PKG; field < CATEGORY; ++field)
	{
		line += strcspn(line, FIELDSEPARATOR) + 1;
	}

	/* OK, now we're pointing at the first character of the CATEGORY field;
		split off all but the first category name */

	len = strcspn(line, CATSEPARATOR);	/* split on space or FIELDSEPARATOR */
	strncpy(category, line, len);	/* copy up to the space or FIELDSEPARATOR */
	category[len] = '\0';			/* and nul-terminate the category */

	return category;
}

/*****************************************************************************/

static int validate_index_line(const char *line)
/*
	Make sure all of the fields we need in an index line actually contain data

	Returns 1 if line is valid, 0 if any of the tested fields are empty
 */
{
	split_index_line(line);

	if ((strlen(index_field[PKG]) == 0) ||
		(strlen(index_field[PORT]) == 0) ||
		(strlen(index_field[COMMENT]) == 0) ||
		(strlen(index_field[DESCR]) == 0) ||
		(strlen(index_field[EMAIL]) == 0) ||
		(strlen(index_field[CATEGORY]) == 0))
		return 0;

	return 1;
}

/*****************************************************************************/

int set_index_categories(void)
/*
	setup an array of pointers, cross-indexed against the categories list,
	to the first line of each category's section within the index buffer.

	this will allow us to easily find the start of a category within the index
	buffer by name, and speed up port searches quite dramatically by limiting
	our searches to only the category to which a given port	belongs.

	on entry, the index buffer will contain a series of lines, each terminated
	with a newline character, with a single nul end-of-file marker

	this routine will replace each newline with a nul character, transforming
	the buffer into an array of strings, with a single extra nul end-of-file
	marker at the end

	each index line will be validated as well, to ensure all of the needed
	fields actually contain data, returning an error if an improperly formatted
	line is encountered

	returns 0 on success, -1 on error
*/
{
	char   *index_ptr = index_buf;		/* pointer into index_buf */
	char	this_category[MAXCATNAME];	/* current category */
	char	next_category[MAXCATNAME];	/* next category */
	int		i = 0;					/* index into index_categories[] array */
	size_t	len;					/* length of current index line */
	int		line_no = 1;			/* track line numbers as we go, in case we
										need to report an improperly formatted
										index line
									 */

	/* allocate memory for the array where we'll save pointers to the first line
		of each category's section within the index buffer
	 */
	if ((index_categories = calloc((num_categories + 1),
									sizeof(char *))) == NULL)
	{
		perror("calloc() failure");
		return -1;
	}

	/* scan the index buffer one newline-terminated line at a time,
		replacing the newline with nul.
		whenever a new category is found, add a pointer to the line
		to the index_categories[] array
	 */
	while (*index_ptr)
	{
		/* whenever we're at the top of this loop, index_ptr will be pointing
			at the start of a new category section in the index buffer (or end
			of buffer)
		 */
		/* replace newline with nul */
		index_ptr[len = strcspn(index_ptr, "\n")]= '\0';

		/* Make sure the line is properly formatted */
		if (!validate_index_line(index_ptr))
		{
			fprintf(stderr, "Error in index file line %d: line = \"%s\"\n",
				line_no, index_ptr);
			return -1;
		}

		(*index_categories)[i++] = index_ptr;	/* save a pointer to this
													category's section
												 */
		/* on all but the first pass, next_category will already contain
			the name of the new category, so we can avoid a needless
			duplication of effort here */
		if (index_ptr == index_buf)
			/* copy this line's category into this_category */
			get_index_line_category(this_category, index_ptr);
		else
			/* copy the category we already picked up on the last pass */
			strcpy(this_category, next_category);

		/* now keep bumping index_ptr to the next index line until we hit a
			new category, or reach end of file
		 */
		while (index_ptr += len + 1)
		{
			if (*index_ptr == '\0')
				break;					/* we've reached end of buffer */

			++line_no;					/* bump the line number counter */

			/* replace newline with nul */
			index_ptr[len = strcspn(index_ptr, "\n")]= '\0';

			/* Make sure the line is properly formatted */
			if (!validate_index_line(index_ptr))
			{
				fprintf(stderr, "Error in index file line %d: line = \"%s\"\n",
					line_no, index_ptr);
				return -1;
			}

			/* get this line's category and compare it to current category */
			get_index_line_category(next_category, index_ptr);

			if (strcmp(this_category, next_category) != 0)
				break;	/* we've hit a new category,
							back to top of outer loop */
		}
	}

	(*index_categories)[i] = NULL;	/* mark end of list */

	return 0;
}

/*****************************************************************************/

static int gen_index_path(char *index_path)
/*
 *	generate the fully qualified pathname to the index file,
 *	storing it back into index_path[]
 *
 *	returns 0 on success, or -1 on error
 */
{
	struct	utsname	uname_struct;		/* information returned by uname() */

	/* get the OS release number */
	if (uname(&uname_struct) == -1)
	{
		perror("uname()");
		return -1;
	}

	/* set index_path to ports_dir/INDEXNAME */
	strcpy(index_path, ports_dir);
	strcat(index_path, INDEXNAME);	/* index name minus OS release number */

	/* append release number to pathname */
	strncat(index_path, uname_struct.release,
		strspn(uname_struct.release, "0123456789"));

	return 0;
}

/*****************************************************************************/

int load_index(void)
/*
	Load the index file into a dynamically allocated buffer which will be used
	throughout the lifetime of the program to ensure the fastest possible access
	to the file's data.

	A pointer to the file's location in memory will be stored in the global
	variable index_buf.

	Returns 0 on success, or -1 on failure
*/
{
	char	index_path[PATH_MAX];

	/* get the fully qualified pathname into index_path */
	if (gen_index_path(index_path) != 0)
		return -1;

	/* load the file into memory, saving its address in the global index_buf */
	if ((index_buf = load_file(index_path)) == NULL)
		return -1;

	return 0;
}
